-- Copyright (c) 2008 Stefan Scheffler

-- Permission is hereby granted, free of charge, to any person
-- obtaining a copy of this software and associated documentation
-- files (the "Software"), to deal in the Software without
-- restriction, including without limitation the rights to use,
-- copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following
-- conditions:

-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
-- OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-- NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
-- HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-- FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
-- OTHER DEALINGS IN THE SOFTWARE.

-- gzip decoder

local bit = dofile("bit.lua")
local luainflate = dofile("luainflate.lua")

local DEBUG = false

local function debug_print(...)
	if DEBUG then
		print(...)
	end
end

local luagzip = {}

function luagzip.new()
	local gzip = {}
	
	local buffer = ""
	local got_header = false
	local inflate = luainflate.new()

	function gzip:decompress(inbuf)
		assert(type(inbuf) == "string", "expected string got "..type(inbuf))

		-- parse header
		if not got_header then
			buffer = buffer..inbuf
			
			local buffersize = #buffer
			-- not enough data to parse the header
			if buffersize < 10 then
				return ""
			end
			
			-- 2 byte signature
			local GZIP_SIG = string.char(0x1f, 0x8b)
			if buffer:sub(1,2) ~= GZIP_SIG then
				return nil, "not a gzip file"
			end
			debug_print("found gzip signature")

			-- 1 byte compression method
			local CM_DEFLATE = 8
			local cm = buffer:sub(3,3):byte()

			if cm ~= CM_DEFLATE then
				return nil, "unsupported compression method"
			end
			debug_print("supported compression method")

			-- 1 byte flags
			local RESERVED_MASK = 0xE0
			local flags = buffer:sub(4,4):byte()
			
			if bit.band(flags, RESERVED_MASK) ~= 0 then
				return nil, "unsupported flags"
			end
			debug_print("flags ok")
			
			-- 4 byte mtime 
			-- 1 byte extra flags
			-- 1 byte OS

			local blockoffset = 11
			-- optional 2 + len bytes fextra block
			local have_fextra = bit.band(flags, 0x4)
			if have_fextra ~= 0 then
				if buffersize <= blockoffset then
					return ""
				end

				local high = buffer:sub(blockoffset, blockoffset):byte()
				local low = buffer:sub(blockoffset + 1, blockoffset + 1):byte()
				blockoffset = blockoffset + 2 + bit.blshift(high, 8) + low
			end

			local function read_cstring(offset)
				local fname = ""
				local got_fname = false
				for i=blockoffset,buffersize do
					local char = buffer:sub(i,i)
					if char:byte() == 0 then
						got_fname = true
						break
					end
					fname = fname..char
				end
				
				if got_fname then
					return fname
				end
			end

			-- cstring if comment flag set
			local have_fname = bit.band(flags, 0x8)
			if have_fname ~= 0 and buffersize >= blockoffset then
				local str = read_cstring(blockoffset)
				if str then
					debug_print("name: "..str)
				else
					return ""
				end
				blockoffset = blockoffset + #str + 1
			end

			-- cstring if comment flag set
			local have_fcomment = bit.band(flags, 0x10)
			if have_fcomment ~= 0 and buffersize >= blockoffset then
				local str = read_cstring(blockoffset)
				if str then
					debug_print("comment: "..str)
				else
					return ""
				end
				blockoffset = blockoffset + #str + 1
			end

			-- 2 byte crc if crc flag set
			local have_fcrc = bit.band(flags, 0x10)
			if have_fcrc ~= 0 then
				if  buffersize > blockoffset then
					debug_print("got header crc")
					blockoffset = blockoffset + 2
				else
					return ""
				end
			end
			
			-- chop off header
			buffer = buffer:sub(blockoffset)
			got_header = true
		else
			buffer = inbuf
		end

		return inflate:inflate(buffer)
	end
	
	return gzip
end


return luagzip