-- Copyright (c) 2008 Stefan Scheffler

-- Permission is hereby granted, free of charge, to any person
-- obtaining a copy of this software and associated documentation
-- files (the "Software"), to deal in the Software without
-- restriction, including without limitation the rights to use,
-- copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the
-- Software is furnished to do so, subject to the following
-- conditions:

-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
-- OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
-- NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
-- HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
-- WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
-- FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
-- OTHER DEALINGS IN THE SOFTWARE.

-- pure lua deflate decoder
-- based on puff.c by Mark Adler (part of zlib 1.2.3).
-- some ideas for the bit operations taken from luabit

-- TODO: huffman trees as global variables within the instance are kind of lame

local bit = dofile("bit.lua")

local DEBUG = false

local function debug_print(...)
	if DEBUG then
		print(...)
	end
end

local function shorttonum(str)
	local low = str:sub(1, 1):byte()
	local high = str:sub(2, 2):byte()
	return bit.blshift(high, 8) + low
end

-- create huffman tree from code length table
local function create_tree(lengths)
	if not lengths then
		debug_print("create_tree called without symbol table")
		return
	end

	local tree = {}

	-- create counts table
	tree.count = {}
	for i=0,15 do
		tree.count[i] = 0
	end

	-- count lengths	
	for i,v in pairs(lengths) do
		tree.count[v] = tree.count[v] + 1
		
	end
	
	-- no codes. codes with length 0 never show up
	-- if none ever shows up there's no data
	-- could probably skip this
	if tree.count[0] == #lengths + 1 then
		return tree
	end

	-- offset of each code length group
	local offsets = {}
	offsets[1] = 0
	-- could do this with an iterator
	for i=1,14 do
		offsets[i + 1] = offsets[i] + tree.count[i]
	end

	-- create table of codes in incrementing order
	tree.symbol = {}
	for i=0,#lengths do
		local length = lengths[i]
		if length ~= 0 then
			tree.symbol[offsets[length]] = i
			offsets[length] = offsets[length] + 1
		end
	end

	return tree
end

local luainflate = {}

function luainflate.new()
	local deflate = {}
	
	local buffer = "" -- input buffer
	local out = "" -- output buffer

	local bufferpos = 1 -- byte position in buffer. 1 based
	local offset = 0 -- bit offset in byte pointed to by bufferpos. 0 based

	local lastblock = false -- current block is last one in file
	local blocktype = 0 -- type of block

	local got_header = false -- block header of current block successfully parsed

	local got_tables = false -- huffman tables for current compressed block built
	local blocksize = nil -- size of uncompressed block. nil if not determined

	local eof = false -- end of file. set if block marked with lastblock was successfully decompressed

	local ltree
	local dtree


	-- return and clear output buffer
	local function flush()
		local ret = out
		out = ""
		return ret
	end

	local divtable = {2, 4, 8, 16, 32, 64, 128}

	-- load some bits from buffer
	local function get_bits(num)
--		debug_print("get_bits: "..num.." bsize "..#buffer.." pos "..bufferpos.." offset "..offset)

		-- save bufferpos and offset to restore them if it runs out of data
		local oldbufferpos = bufferpos
		local oldoffset = offset

		local ret = 0

		local currentbyte = buffer:sub(bufferpos, bufferpos):byte()

		local exp = 1
		for i=0,num-1 do
			if bufferpos > #buffer then
				debug_print("get_bit: out of input data")
				bufferpos = oldbufferpos
				offset = oldoffset
				debug_print("remains: "..num.." bsize "..#buffer.." pos "..bufferpos.." offset "..offset)
				return
			end

			-- shift to the wanted bit
			local b = currentbyte
			if offset ~= 0 then
				-- division via lookup table
				-- best case is about 3 times faster than looping
				-- worst case is about equal
				b = b / divtable[offset]
			end
			-- determine its value
			b = math.floor(b) % 2

			-- end of byte reached
			if offset == 7 then
				offset = 0
				bufferpos = bufferpos + 1
				currentbyte = buffer:sub(bufferpos, bufferpos):byte()

			else
				offset = offset + 1
			end

			-- XXX: the exp func is probably costing a lot
--			ret = ret + (b * 2 ^ i)
			-- this doesn't save much over the one above ..weird
			ret = ret + (b * exp)
			exp = exp * 2
		end
		return ret
	end

	-- decode a code
	local function decode(tree)
		local code = 0
		local first = 0
		local index = 0

		local currentbyte = buffer:sub(bufferpos, bufferpos):byte()

		for len=1,15 do
			-- inlined get_bits() for 1bit loads
			-- saves a couple of sub() over multiple get_bits()

			-- out of data
			if bufferpos > #buffer then
				debug_print("decode() end of buffer")
				return
			end

			local val = currentbyte
			if offset ~= 0 then
				val = val / divtable[offset]
			end
			val = math.floor(val) % 2
			
			if offset == 7 then
				offset = 0
				bufferpos = bufferpos + 1
				currentbyte = buffer:sub(bufferpos, bufferpos):byte()
				
			else
				offset = offset + 1
			end

			code = code + val -- add a new bit from source to code
			local count = tree.count[len] -- number of codes of this length

			if code < first + count then -- check if the code is in this length group
				--debug_print("decode "..tostring(index + (code - first)))
				return tree.symbol[index + (code - first)] -- absolute index of first element in length group + relative offset
			end

			index = index + count
			first = first + count 

			-- append zeros
			-- multiplying should be save
			code = code * 2 -- code can't be more than 15 bits
			first = first * 2 -- count can't be more than 9 bits + 15 for shifts(?)
		end
	end

	-- buffer for looking up repetitions
	local window = {}
	local window_pos = 1

	local function window_allocate()
		for i=1,32768 do
			window[i] = ""
		end
	end
	window_allocate()

	local function window_insert(item)
		window[window_pos] = item
		window_pos = (window_pos < 32768) and (window_pos + 1) or 1
	end

	local function window_get(idx)
--		assert(idx <= 32768 and idx > 0 , "get ring item: index out of range: "..idx)
		if idx >= window_pos then
			return window[(32768 - (idx - window_pos))]
		else
			return window[window_pos - idx]
		end
	end
	
	-- decode compressed block
	local function codes()
		-- some of these look like they could be generated at runtime
		local lens = {3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31,
			      35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258}
		local lext = {0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2,
			      3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 0}

		local dists = {1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193,
			       257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145,
			       8193, 12289, 16385, 24577}
		local dext = {0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6,
			       7, 7, 8, 8, 9, 9, 10, 10, 11, 11,
			       12, 12, 13, 13}

		local outbuf = ""
		local symbol

		repeat
			-- duplicate the bufferpos and offset and restore them if get_bit fails (end of data)
			local oldbufferpos = bufferpos
			local oldoffset = offset
--			debug_print("codes(): bsize "..#buffer.." pos "..bufferpos.." offset "..offset)

			local symbol = decode(ltree)
--			debug_print("literal/length decode "..tostring(symbol))

			-- out of data
			if not symbol then
				bufferpos = oldbufferpos
				offset = oldoffset
				debug_print("codes() symbol: end of buffer")
				return outbuf

			-- invald code
			-- this can't really ever happen so it could probably be skipped
			elseif symbol < 0 then
				return nil, nil, "invalid code in compressed block"

			elseif symbol < 256 then -- literal
				local decodedsym = string.char(symbol)
--				debug_print("appending sym "..symbol)
				outbuf = outbuf..decodedsym
				window_insert(decodedsym)

			elseif symbol > 256 then -- length code
				symbol = symbol - 257
				-- extra bits of length code
				local bits = get_bits(lext[symbol + 1])
				-- out of data
				if not bits then
					bufferpos = oldbufferpos
					offset = oldoffset
					debug_print("codes() symbol extra: end of buffer")

					return outbuf
				end
				local len = lens[symbol + 1] + bits

				-- decode distance code
				symbol = decode(dtree)
				-- out of data
				if not symbol then
					bufferpos = oldbufferpos
					offset = oldoffset
					debug_print("codes() distance: end of buffer")

					return outbuf
				-- invalid code
				elseif symbol < 0 then 
					return nil, nil, "invalid code in compressed block"
				end

				-- extra bits of distance code
				local bits = get_bits(dext[symbol + 1])
				-- out of data
				if not bits then
					bufferpos = oldbufferpos
					offset = oldoffset
					debug_print("codes() distance extra: end of buffer")
					return outbuf
				end

				local dist = dists[symbol + 1] + bits
				
				-- copy sequences from window according to decoded length
				-- and distance codes
				-- TODO: this would probably be more efficient if it copied a
				-- a whole sequence instead of single bytes
				-- debug_print("codes() len "..tostring(len).." dist "..tostring(dist))
				local buf = ""
				local pos = dist

				while len > 0 do
					buf = buf..window_get(pos)

					if pos <= 1 then
						pos = dist
					else
						pos = pos - 1
					end
					len = len - 1
				end
				
				-- add decoded bytes to window buffer
				for i=1,#buf do
					window_insert(buf:sub(i,i))
				end

				outbuf = outbuf..buf				
			end			
		until symbol == 256

		got_header = false
		got_tables = false

		debug_print("block processed")
		return outbuf, true
	end

	-- compression modes

	local function block_uncompressed()
		if not blocksize then
			-- not enough data
			if #buffer < 5 then
				debug_print("not enough data to get uncompressed block header")
				return
			end
			
			-- discard first byte if not aligned
			if offset ~= 0 then
				-- buffer = buffer:sub(2)
				bufferpos = bufferpos + 1
				offset = 0
			end

			local size = shorttonum(buffer:sub(bufferpos, bufferpos + 1))
			local size_compl = shorttonum(buffer:sub(bufferpos + 2, bufferpos + 3))
			if size ~= 0xffff - size_compl then
				return nil, "invalid block header in uncompressed block"
			end
			debug_print("blocksize: "..tostring(size))
			

			buffer = buffer:sub(bufferpos + 4)
			bufferpos = 1

			blocksize = size
		end

		if blocksize <= #buffer then
--			debug_print("blocksize smaller or equal to buffer")
			-- block is smaller than buffer
			-- return block sized chunk of buffer
			-- and look for new header
			local buf = buffer:sub(bufferpos, bufferpos + blocksize - 1)
			for i=1,#buf do
				window_insert(buf:sub(i,i))
			end
			out = out..buf

			buffer = buffer:sub(bufferpos + blocksize)
			bufferpos = 1

			blocksize = nil
			got_header = false
			return true
		else
--			debug_print("blocksize larger than buffer")
			-- block is larger than buffer
			-- return the whole buffer
			for i=1,#buffer do
				window_insert(buffer:sub(i,i))
			end
			out = out..buffer
			blocksize = blocksize - #buffer

			bufferpos = 1
			buffer = ""
		end
	end
	
	
	local function block_fixed()
		if not got_tables then
			-- generate table from hardcoded values
			local llengths = {}
			for i=0,287 do
				if i < 144 then
					llengths[i] = 8
				elseif i < 256 then
					llengths[i] = 9
				elseif i < 280 then
					llengths[i] = 7
				else
					llengths[i] = 8
				end
			end

			ltree = create_tree(llengths)

			local dlengths = {}
			for i=0,29 do
				dlengths[i] = 5
			end
			dtree = create_tree(dlengths)
			got_tables = true
		end

		local buf, complete, err = codes()
		if not err then
			out = out..buf
			buffer = buffer:sub(bufferpos)
			bufferpos = 1
		end
		
		return complete, err
	end
	
	local function block_dynamic()
		if not got_tables then
			local oldbufferpos = bufferpos
			local oldoffset = offset
			
			local order = {16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15}
			
			if (offset < 2 and #buffer < 3) or #buffer < 4 then
				debug_print("not enough data to decode dynamic block header")
				return
			end

			-- codelist sizes
			local nlen = get_bits(5) + 257
			local ndist = get_bits(5) + 1
			local ncode = get_bits(4) + 4

			-- TODO verify
--			debug_print("nlen "..tostring(nlen))
--			debug_print("ndist "..tostring(ndist))
--			debug_print("ncode "..tostring(ncode))

			local lengths = {}

			-- decompress header			
			-- read the ncode 3 bit codes and put store them in predefined oreder
			for i=1,ncode do
				-- not enough data
				local bits = get_bits(3)
				if not bits then
					bufferpos = oldbufferpos
					offset = oldoffset
					return
				end
				lengths[order[i]] = bits
			end

			for i=ncode+1,19 do
				lengths[order[i]] = 0
			end

			for i=0,18 do
				debug_print("len: "..i.." "..tostring(lengths[i]))
			end

			local cltree = create_tree(lengths)
			-- XXX it's assuming that it doesn't have data
			-- no cltree could also mean that the header is corrupted
			if not cltree then
				debug_print("failed to create code length code tree")
				
				bufferpos = oldbufferpos
				offset = oldoffset
				return
			end
--			debug_print("cltree created successfully")

			local function get_lengths(num)
				local lengths = {}
				local index = 0
				while index < num do
--					debug_print("index "..index)
					local symbol = decode(cltree)
					if not symbol then
						return
					elseif symbol < 16 then -- literal length
						lengths[index] = symbol
						index = index + 1
					else -- run length code
						-- determine actual repetition type
						local len = 0
						-- repeat previous 3-6 times
						if symbol == 16 then
							if index == 0 then 
								return
							end
							-- previous length
							len = lengths[index - 1]

							local bits = get_bits(2)
							if not bits then
								return
							end
							symbol = 3 + bits
							-- fill with given number of 0s (3-10) 
						elseif symbol == 17 then
							local bits = get_bits(3)
							if not bits then
								return
							end
							symbol = 3 + bits
							-- fill with given number of 0s (11-something)
						else
							local bits = get_bits(7)
							if not bits then
								return
							end
							symbol = 11 + bits
						end

						if index + symbol > nlen + ndist then
							return
						end
--						debug_print("duplicating: "..symbol)
						while symbol > 0 do
							lengths[index] = len
							symbol = symbol - 1
							index = index + 1
						end
					end
				end
				return lengths
			end

--			debug_print("create ltree")
			ltree = create_tree(get_lengths(nlen))
--			debug_print("create dtree")
			dtree = create_tree(get_lengths(ndist))

			-- didn't get the whole header
			if not (ltree and dtree) then
				bufferpos = oldbufferpos
				offset = oldoffset
				return
			end

			got_tables = true
		end

		local buf, complete, err = codes()
		if not err then
			out = out..buf
			buffer = buffer:sub(bufferpos)
			bufferpos = 1
		end
--		print("dynamic block complete "..tostring(complete))

		return complete, err
	end

	function deflate:inflate(inbuf)
--		assert(type(buffer) == "string")	
--		debug_print("new chunk obsize "..#buffer.." nbsize "..#inbuf.." offset "..offset)

		if eof then
			return
		end

		buffer = buffer..inbuf
		
		-- have a complete block
		local complete = true

		-- decode blocks till arriving at an incomplete one
		repeat
			-- parse header
			if not got_header then
				local buffersize = #buffer

				-- check if there are enough bits for header in buffer (3 bits)
				-- XXX just buffersize > 1 would probably do too
				if (buffersize > 0 and offset <= 5) or buffersize > 1 then
					-- read final block bit
					-- buffer size verfied above
					local val = get_bits(1)
					if val == 1 then
						lastblock = true
					end
					debug_print("last block "..tostring(lastblock))
					
					-- get block type
					-- buffer size verified above
					blocktype = get_bits(2)
					if blocktype == 3 then
						return nil, "invalid block type"
					end
					debug_print("block type: "..blocktype)
					got_header = true
				else
					break
				end
			end
			
			if got_header then
				debug_print("decoding data block")

				-- decode with correct decoding function
				local err

				if blocktype == 0 then
					complete, err = block_uncompressed()
				elseif blocktype == 1 then
					complete, err = block_fixed()
				elseif blocktype == 2 then
					complete, err = block_dynamic()
				else
					return nil, "invalid block type"
				end
				
				-- return error and abort processing
				if err then
					return nil, err
				elseif lastblock and complete then
					eof = true
					break
				end
			end
			
		until not complete

		return flush()
	end
	
	return deflate
end


return luainflate